/***************************************************************************
                               ctelnet.cpp
                             -------------------
    begin                : Mon May 21 2001
    copyright            : (C) 2001 by Kmud Developer Team
    email                : kmud-devel@kmud.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ctelnet.h"

#define TELCMDS 1
#define TELOPTS 1
#include <arpa/telnet.h>

#include <qtimer.h>
#include <kapp.h>
#include <kmuddebug.h>

#include "cmccp.h"

// not defined in telnet standard.
#define TELOPT_COMPRESS		85
#define TELOPT_COMPRESS2	86
#define TELOPT_MSP		90
#define TELOPT_MXP		91

// defines for telnet negotiation states
#define TELNEG_STATE_NORMAL     0	// normal: no IAC received
#define TELNEG_STATE_IAC        1	// IAC received, expecting telnet command
#define TELNEG_STATE_IAC_SB     2	// IAC + SB, we're in subnegotiations
#define TELNEG_STATE_IAC_WILL   3	// IAC + telnet command received, expecting telnet option
#define TELNEG_STATE_IAC_WONT   4	// IAC + telnet command received, expecting telnet option
#define TELNEG_STATE_IAC_DO     5	// IAC + telnet command received, expecting telnet option
#define TELNEG_STATE_IAC_DONT   6	// IAC + telnet command received, expecting telnet option

CTelnet::CTelnet(QObject* parent, const char* name)
:QObject(parent,name)
{
	m_echo=true;
	m_buffer.resize(0);
	m_socket = new QSocket(this,"CTelnet::socket");
	QObject::connect(m_socket,SIGNAL(error(int)),this,SLOT(socketErrorReceived(int)));
	setFilterName("Telnet");

	m_mccp = new CMCCP;
	QObject::connect(m_mccp, SIGNAL(sendToServer(QString)), this, SLOT(sendData(QString)));
}

CTelnet::~CTelnet()
{
	if (isConnected())
	{
		close();
	}
	QObject::disconnect(m_socket,SIGNAL(error(int)),this,SLOT(socketErrorReceived(int)));

	if (m_mccp != 0)
		delete m_mccp;
}


int CTelnet::connect(QString hostname, unsigned long port, unsigned int timeout)
{

	KApplication* myApp = KApplication::kApplication();
	int retVal=0;
	
	kdDebug(DAREA_CORE_TELNET) << "connecting to " << hostname << ":" << port << " (timeout: " << timeout << ")" << endl;
	
	m_socket->connectToHost(hostname, port);

	m_timerEventCalled=false;
	m_socketErrorOccured=false;
	
	QObject::startTimer(timeout*1000);
	
	while ( (m_socket->state() != QSocket::Connection)
					&& (m_timerEventCalled == false)
					&& (m_socketErrorOccured == false ) )
	{
		myApp->processEvents();
	}
	
	QObject::killTimers();

	if (m_socket->state() != QSocket::Connection)
	{
		retVal=-255;
		if (m_timerEventCalled==true) retVal=-3; // connection timeout
		if (m_socketErrorOccured==true)
		{
			if (m_socketError==QSocket::ErrConnectionRefused) retVal=-2;
			if (m_socketError==QSocket::ErrHostNotFound) retVal=-1;
		}
	}

	kdDebug(DAREA_CORE_TELNET) << "connect result: " << retVal << endl;	
	
	if (retVal==0) // if seccessfully connected
	{
		m_telnegState = TELNEG_STATE_NORMAL;
		m_telnegDebugOutput = "";
		m_subnegs.resize(0);
		setEcho(true); // no telnet echo -> local echo. this is default (says RFC), do not change!
		QObject::connect(m_socket,SIGNAL(connectionClosed()),this,SLOT(close()));
		QObject::connect(m_socket,SIGNAL(readyRead()),this,SLOT(receiveData()));

		// if we missed a readyRead signal while doing processEvent
		// simulate it by using a single shot timer
		if (m_socket->bytesAvailable() > 0)
			QTimer::singleShot(0, this, SLOT(receiveData()));
	}
	

	
	return retVal;
}


void CTelnet::close()
{
	QObject::disconnect(m_socket,SIGNAL(connectionClosed()),this,SLOT(close()));
	QObject::disconnect(m_socket,SIGNAL(readyRead()),this,SLOT(receiveData()));
	m_socket->close();
	m_telnegState = TELNEG_STATE_NORMAL;
	m_mccp->reset();
	emit(connectionClosed());
	kdDebug(DAREA_CORE_TELNET) << "Connection closed." << endl;
}

	
bool CTelnet::isEchoOn()
{
	return m_echo;
}

	
void CTelnet::setEcho(bool b)
{
	m_echo=b;
	emit changedEcho(b);
}


QString CTelnet::getHostName() const
{
	return m_socket->peerName();
}

QString CTelnet::getHostIP() const
{
	return m_socket->peerAddress().toString();
}

unsigned long CTelnet::getHostPort() const
{
	return m_socket->peerPort();
}

bool CTelnet::isConnected() const
{
	return (m_socket->state() == QSocket::Connection);
}



int CTelnet::receiveData()
{
	// FIXME_stephan glue open ANSI commands splitted in two transmissions
	
	int bytesRead, bytesTelnet;
	QString s;

	m_buffer = m_socket->readAll();
	m_mccp->process(m_buffer);
	if (m_buffer.size() == 0) return 0;

	bytesRead = m_buffer.size();
	bytesTelnet = processCommands();
	
	// add a \0 every time when converting to QString
	// QByteArray::data() is not null terminated
	m_buffer.resize(m_buffer.size()+1);
	m_buffer[m_buffer.size()-1]='\0';
	s = m_buffer.data();
	m_buffer.resize(m_buffer.size()-1);	
	
	kdDebug(DAREA_CORE_TELNET) << "read " << bytesRead << " bytes, (" << bytesTelnet << " Telnet)" << endl;
	emit(filledBuffer(s));
	
	return bytesRead;
}

int CTelnet::sendData(QString _buf) const
{
	int bytesToWrite=0, bytesWritten=0;
	QString buf=_buf;
	
	bytesToWrite=buf.length();
		
	bytesWritten = m_socket->writeBlock(buf.latin1(), bytesToWrite);

	kdDebug(DAREA_CORE_TELNET) << "wrote " << bytesWritten << "/" << bytesToWrite << " Bytes" << endl;
	
	return bytesWritten;
}

bool CTelnet::isCommand(unsigned char c) const
{
	return TELCMD_OK(c);
}

bool CTelnet::isOption(unsigned char c) const
{
	if ((c==TELOPT_COMPRESS) || (c==TELOPT_COMPRESS2) || (c==TELOPT_MSP) || (c==TELOPT_MXP))
		return true;

	return TELOPT_OK(c);
}

QString CTelnet::codeToString(unsigned char c) const
{
	QString s;
	
	if (c == TELOPT_COMPRESS) s = "COMPRESS";
	else
	if (c == TELOPT_COMPRESS2) s = "COMPRESS2";
	else
	if (c == TELOPT_MSP) s = "MSP";
	else
	if (c == TELOPT_MXP) s = "MXP";
	else
	if (isCommand(c)) s = QString(TELCMD(c));
	else
	if (isOption(c)) s = QString(TELOPT(c));
	else
	s.setNum((int)c);
	
	return s;
}

int CTelnet::processCommands()
{
	unsigned char c;		// current byte from the buffer
	int i;				// current pos in buffer
	bool isTelneg;			// true when byte belongs to a telneg sequence
	bool sendAnswer;		// true when we answer to telneg requests
					// not always true to prevent nonterminating loops, see RFC
	QByteArray newBuffer;		// new buffer without telnet codes
	QString	s;			// temp. string
	QString sendBuf;		// buffer sent as answer to telnegs received

	
	newBuffer.resize(0);

	for (i=0;i<(int)m_buffer.size();i++)
	{
		c = (unsigned char)m_buffer[i];
		isTelneg = false;
		switch (m_telnegState)
		{
			case TELNEG_STATE_NORMAL:
				if (c == IAC)
				{
					m_telnegState = TELNEG_STATE_IAC;
					m_telnegDebugOutput = codeToString(c);
					isTelneg = true;
				}
				break;

			case TELNEG_STATE_IAC:
				if (isCommand(c))
				{
					m_telnegDebugOutput += "_" + codeToString(c);
					isTelneg = true;
					switch (c)
					{
						case SB: m_telnegState = TELNEG_STATE_IAC_SB; break;
						case DO: m_telnegState = TELNEG_STATE_IAC_DO; break;
						case DONT: m_telnegState = TELNEG_STATE_IAC_DONT; break;
						case WILL: m_telnegState = TELNEG_STATE_IAC_WILL; break;
						case WONT: m_telnegState = TELNEG_STATE_IAC_WONT; break;
						default:
							{
								kdDebug(DAREA_CORE_TELNET) << "Unhandled Telnet command: " << m_telnegDebugOutput << endl;
								m_telnegState = TELNEG_STATE_NORMAL; // reset to normal
							}
					}
				}
				else
				{
					m_telnegDebugOutput += "_" + codeToString(c); 
					isTelneg = true;
					kdDebug(DAREA_CORE_TELNET) << "Unknown Telnet command: " << m_telnegDebugOutput << endl;
					m_telnegState = TELNEG_STATE_NORMAL; // reset to normal
				}
				break;
				
			case TELNEG_STATE_IAC_SB:
				m_telnegDebugOutput += "_" + codeToString(c);
				if (c == SE)
				{
					kdDebug(DAREA_CORE_TELNET) << "Received: " << m_telnegDebugOutput << endl;
					m_telnegDebugOutput="";

					// subnegotiation handling here

					m_subnegs.resize(0);
					m_telnegState = TELNEG_STATE_NORMAL;
				}
				else
				{
					m_subnegs.resize(m_subnegs.size()+1);
					m_subnegs[m_subnegs.size()-1]=c;
				}
				isTelneg = true;
				break;

			case TELNEG_STATE_IAC_WILL:
			case TELNEG_STATE_IAC_WONT:
			case TELNEG_STATE_IAC_DO:
			case TELNEG_STATE_IAC_DONT:
				m_telnegDebugOutput += "_" + codeToString(c);
				isTelneg = true;
				if (isOption(c))
				{
					kdDebug(DAREA_CORE_TELNET) << "Received: " << m_telnegDebugOutput << endl;
					sendBuf="";
					sendBuf[0]=IAC;
					sendAnswer = true;

					switch (c)
					{
						// echo handling
						case TELOPT_ECHO:
							switch (m_telnegState)
							{
								case TELNEG_STATE_IAC_WILL: sendBuf[1]=DO; setEcho(false); break;
								case TELNEG_STATE_IAC_DO: sendBuf[1]=WILL; setEcho(true); break;
								case TELNEG_STATE_IAC_WONT: sendBuf[1]=DONT; setEcho(true); break;
								case TELNEG_STATE_IAC_DONT: sendBuf[1]=WONT; setEcho(false); break;
								default: sendAnswer=false;
							}
							sendBuf[2]=TELOPT_ECHO;
							break;
						// reject any other options
						default:
							switch (m_telnegState)
							{
								case TELNEG_STATE_IAC_WILL: sendBuf[1]=DONT; break;
								case TELNEG_STATE_IAC_DO: sendBuf[1]=WONT; break;
								default: sendAnswer=false;
							}
							sendBuf[2]=c;
					}

					if (sendAnswer)
					{
						m_telnegDebugOutput = codeToString((QChar)sendBuf.at(0));
						m_telnegDebugOutput += "_" + codeToString((QChar)sendBuf.at(1));
						m_telnegDebugOutput += "_" + codeToString((QChar)sendBuf.at(2));
						kdDebug(DAREA_CORE_TELNET) << "Answering: " << m_telnegDebugOutput << endl;
						sendData(sendBuf);
					}

				}
				else
				{
					kdDebug(DAREA_CORE_TELNET) << "Unknown Telnet option: " << m_telnegDebugOutput << endl;
				}
				m_telnegDebugOutput="";
				m_telnegState = TELNEG_STATE_NORMAL;
				break;
		}

		if (isTelneg == false)
		{
			// Qt really should have a QByteArray::append()
			newBuffer.resize(newBuffer.size()+1);
			newBuffer[newBuffer.size()-1]=c;
		}
	}

	i = m_buffer.size() - newBuffer.size();
	m_buffer = newBuffer;
	return i;
}

void CTelnet::timerEvent(QTimerEvent* t)
{
	m_timerEventCalled=true;
}

void CTelnet::socketErrorReceived(int e)
{
	m_socketErrorOccured=true;
	m_socketError=(QSocket::Error)e;
}

QString CTelnet::filterInput(QString& _s)
{
	this->sendData(_s);
	return _s;
}

