/***************************************************************************
                               kmudconfigmanager.cpp
                             -------------------
    begin                : Thu Aug 23 2001
    copyright            : (C) 2001 by Kmud Developer Team
    email                : kmud-devel@kmud.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "kmudconfigmanager.h"

// Qt inlcudes
#include <qdom.h>
#include <qfileinfo.h>
#include <qfile.h>
#include <qdir.h>

// KDE includes
#include <kstddirs.h>
#include <kconfig.h>
#include <kapp.h>

// Kmud includes
#include "kmuddebug.h"
#include "kmudconfigcache.h"
#include "cmudprofile.h"

///////////////////////////////////////////////////////////////////////////////

/**
  * Private class implementing the ConfigAPI interface.
  * Creating of it is handled by KmudConfigManager.
  */
class ConfigAPIImpl : public ConfigAPI {
public:
	ConfigAPIImpl() {
		m_configCache = new KmudConfigCache();
	}
	~ConfigAPIImpl() {};
	
	KConfig* kmudConfig() { return kapp->config(); }

	KmudConfigCache* configCache() { return m_configCache; }

private:
	KmudConfigCache* m_configCache;
};


///////////////////////////////////////////////////////////////////////////////

KmudConfigManager* KmudConfigManager::m_self = 0;

///////////////////////////////////////////////////////////////////////////////

KmudConfigManager* KmudConfigManager::self()
{
	if (m_self == 0)
	{
		m_self = new KmudConfigManager();
		m_self->configCache()->load(); // have to do this after m_self is set as it uses self()
	}
	
	return m_self;
}

///////////////////////////////////////////////////////////////////////////////

KmudConfigManager::KmudConfigManager()
{
	m_configAPI = new ConfigAPIImpl;
	QObject::connect(m_configAPI, SIGNAL(configGroupChanged(const QString&)),
	                 this, SLOT(slotConfigDataChanged(const QString&)));
	
	m_mudProfiles.setAutoDelete(true);
	QString globalKmudDir = locate("appdata", "kmudui.rc");
	if (!globalKmudDir.isEmpty())
	{
		QFileInfo fileInfo(globalKmudDir);
		globalKmudDir = fileInfo.dirPath(true);
		globalKmudDir.append("/muds/");
		m_paths.append(globalKmudDir);
	}
	m_defaultPath = kapp->dirs()->saveLocation("appdata", "muds/");
	m_paths.prepend(m_defaultPath);
}

///////////////////////////////////////////////////////////////////////////////

KmudConfigManager::~KmudConfigManager()
{
}

///////////////////////////////////////////////////////////////////////////////

ConfigAPI* KmudConfigManager::configAPI()
{
	return m_configAPI;
}

///////////////////////////////////////////////////////////////////////////////

KmudConfigCache* KmudConfigManager::configCache()
{
	return m_configAPI->configCache();
}

///////////////////////////////////////////////////////////////////////////////

QDomElement KmudConfigManager::openProfile(const QString& name, bool create)
{
	QStringList names = QStringList::split("/", name);
	
	if (names.count() < 1 || names.count() > 2)
		return QDomElement();
	
	CMudProfile* mudProfile = m_mudProfiles[names[0]];
	if (mudProfile == 0)
	{
		mudProfile = loadMudProfile(names[0]);
		if (mudProfile != 0)
		{
			m_mudProfiles.insert(names[0], mudProfile);
		}
		else if (create)
		{
			mudProfile = new CMudProfile;
			mudProfile->create();
			m_mudProfiles.insert(names[0], mudProfile);
		}
		else
			return QDomElement();
	}
	
	// requested profile is a mud profile
	if (names.count() == 1)
	{
		return mudProfile->profileElement();
	}
	
	CProfile* charProfile = mudProfile->charProfile(names[1]);
	if (charProfile == 0)
	{
		charProfile = loadCharProfile(names[0], names[1]);
		if (charProfile != 0)
		{
			mudProfile->addCharProfile(names[1], charProfile);
		}
		else if (create)
		{
			charProfile = new CProfile;
			charProfile->create();
			mudProfile->addCharProfile(names[1], charProfile);
		}
		else
			return QDomElement();
	}
	
	return charProfile->profileElement();
}

///////////////////////////////////////////////////////////////////////////////

bool KmudConfigManager::saveProfile(const QString& name)
{
	kdDebug(DAREA_CORE_CONFIGMANAGER) << "saveProfile(" << name << ")\n";
	QStringList names = QStringList::split("/", name);
	if (names.count() < 1 || names.count() > 2)
		return false;
	
	QString mudName = names[0];
	QString charName = (names.count() == 1 ? QString::null : names[1]);

	CMudProfile* mudProfile = m_mudProfiles[mudName];
	if (mudProfile == 0)
		return false;
	
	// construct save path
	QString path = profileSaveFilename(mudName, charName);
	
	if (names.count() == 1)
	{
		return mudProfile->saveAs(path);
	}
	
	CProfile* charProfile = mudProfile->charProfile(charName);
	if (charProfile == 0)
		return false;
	
	return charProfile->saveAs(path);
}

///////////////////////////////////////////////////////////////////////////////

void KmudConfigManager::setDefaultPath(const QString& defaultPath)
{
	if (defaultPath.isEmpty())
		return;
	
	QFileInfo fileInfo(defaultPath);
	if (fileInfo.isDir() && fileInfo.isWritable())
	{
		m_defaultPath = fileInfo.absFilePath();
		if (m_defaultPath.right(1) != "/")
			m_defaultPath.append("/");
	}
	m_paths.prepend(m_defaultPath);
}

///////////////////////////////////////////////////////////////////////////////

QStringList KmudConfigManager::listProfiles(const QString& baseProfile) const
{
	QStringList result;
	QStringList names;
	if (!baseProfile.isEmpty())
		 names = QStringList::split("/", baseProfile);

	if (names.count() > 1)
		return result;

	if (names.count() == 0) // looking for mudprofiles
	{
		for (uint i = 0; i < m_paths.count(); ++i)
		{
			QDir dir(m_paths[i]);
			QStringList entries = dir.entryList();
			for (uint j = 0; j < entries.count(); ++j)
			{
				QString fileName = entries[j];
				entries[j].prepend("/");
				entries[j].prepend(m_paths[i]);
				QFileInfo fileInfo(entries[j]);

				if (fileInfo.isDir())
				{
					QFileInfo fileInfo2(entries[j] + "/" + fileName + ".kmudprf");
					if (fileInfo2.exists() && fileInfo2.isFile() && fileInfo2.isReadable())
					{
						if (!result.contains(fileName));
							result.append(fileName);
					}
				}
			}
		}
	}
	else
	{
		// list character profiles of names[0];
		QString encodedMudName = QFile::encodeName(names[0]);
		for (uint i = 0; i < m_paths.count(); ++i)
		{
			QString path = m_paths[i];
			path.append(encodedMudName);
			path.append("/characters");

			QStringList profiles = listProfileFiles(path);
			for (uint j = 0; j < profiles.count(); ++j)
			{
				QString profile = profiles[j].left(profiles[j].findRev("."));
				if (!result.contains(profile))
					result.append(profile);
			}
		}
	}
	return result;
}

///////////////////////////////////////////////////////////////////////////////

QStringList KmudConfigManager::listProfileDirs(const QString& mudProfile) const
{
	QStringList result;
	if (mudProfile.isEmpty())
		return result;

	QString encodedMudName = QFile::encodeName(mudProfile);
	for (uint i = 0; i < m_paths.count(); ++i)
	{
		QString path = m_paths[0];
		path.append("/");
		path.append(encodedMudName);
		QFileInfo fileInfo(path);
		if (fileInfo.exists() && fileInfo.isDir())
		{
			path.append("/");
			result.append(path);
		}
	}

	return result;
}

///////////////////////////////////////////////////////////////////////////////

QString KmudConfigManager::defaultSavePath(const QString& mudProfile)
{
	if (mudProfile.isEmpty())
		return QString::null;

	QString result = profileSaveFilename(mudProfile);
	int filenamePos = result.findRev("/");
	if (filenamePos > 0)
	{
		result = result.left(filenamePos+1);
	}
	
	return result;
}

///////////////////////////////////////////////////////////////////////////////

CMudProfile* KmudConfigManager::loadMudProfile(const QString& mudName)
{
	QStringList paths = findProfile(mudName);
	if (paths.count() == 0)
		return 0;
	
	CMudProfile* profile = new CMudProfile();
	if (profile->open(paths[paths.count()-1]))
		return profile;
	
	delete profile;
	return 0;
}

///////////////////////////////////////////////////////////////////////////////

CProfile* KmudConfigManager::loadCharProfile(const QString& mudName, const QString& charName)
{
	QStringList paths = findProfile(mudName, charName);
	if (paths.count() == 0)
		return 0;

	CProfile* profile = new CProfile();
	if (profile->open(paths[0]))
		return profile;
	
	delete profile;
	return 0;
}

///////////////////////////////////////////////////////////////////////////////

QString KmudConfigManager::profileFilename(const QString& mudName, const QString& charName) const
{
	QString encodedMudName = QFile::encodeName(mudName);
	QString profile = encodedMudName;
	profile.append("/");
	if (charName.isEmpty())
	{
		profile.append(encodedMudName);
	}
	else
	{
		profile.append("characters/");
		profile.append(QFile::encodeName(charName));
	}
	profile.append(".kmudprf");

	return profile;
}

///////////////////////////////////////////////////////////////////////////////

QStringList KmudConfigManager::findProfile(const QString& mudName, const QString& charName) const
{
	QStringList resultPaths;
	QString profile = profileFilename(mudName, charName);
	
	for (uint i = 0; i < m_paths.count(); ++i)
	{
		QString fullPath = m_paths[i];
		fullPath.append(profile);
		
		QFileInfo pathInfo(fullPath);
		if (pathInfo.exists())
		{
			resultPaths.append(fullPath);
		}
	}
	
	return resultPaths;
}

///////////////////////////////////////////////////////////////////////////////

QString KmudConfigManager::profileSaveFilename(const QString& mudName, const QString& charName)
{
	QString fullPath = m_defaultPath;
	fullPath.append(profileFilename(mudName, charName));

	int filenamePos = fullPath.findRev("/");
	if (filenamePos > 0)
	{
		KStandardDirs::makeDir(fullPath.left(filenamePos));
	}
	
	return fullPath;
}

///////////////////////////////////////////////////////////////////////////////

QStringList KmudConfigManager::listProfileFiles(const QString& directory) const
{
	if (directory.isEmpty())
		return QStringList();

	QDir dir(directory, "*.kmudprf");

	if (dir.exists() && dir.isReadable())
		return dir.entryList();
	else
		return QStringList();
}

///////////////////////////////////////////////////////////////////////////////

void KmudConfigManager::slotConfigDataChanged(const QString& group)
{
	// FIXME_kevin: update only if groups cached by the ConfigCache have changes
	configCache()->load();
}

// end of file
